#include "synth.h"

Synth::Synth(double sampleRate){
    setSampleRate(48000.0, sampleRate);
}

#define frac(x) ((x) - ((long)x))
constexpr double MIN_VOLUME = 0.00001;

static constexpr double amin = 50;
static constexpr double apiq = 500;
static constexpr double amax = 10000;
//Volume of voice as a function of sample rate independent frequency.
static constexpr double getAmplitude( double hz )
{
  if( hz < amin )       return 0.0;
  if( hz < apiq )      {
        double a = (hz - amin) / (apiq - amin);
        return a * a;}
  if( hz < amax )     {
        double a  = 1.0 - (hz - apiq) / (amax - apiq);
        return a * a;
  }
  return 0.0;
}

//Sanity checks: voices should become silent outside audible frequencies.
//Voice index code will loop infinitely if these fail.
static_assert( MIN_VOLUME > getAmplitude(10.0) );
static_assert( MIN_VOLUME > getAmplitude(20000.0));

//New fundamental is twice as high as old fundamental.
//Even overtones become plain overtones.
void Synth::shiftUp()
{
    spectrumPhase += 1.0;
    for(uint voice = 0; voice < NUM_VOICES; ++voice){
        phases[voice] = phases[(2 * voice + 1) & VOICE_MASK];
    }
}

//New fundamental is half as high as old fundamental.
//Overtones become even overtones.
void Synth::shiftDown()
{
    spectrumPhase -= 1.0;
    for(uint voice = NUM_VOICES / 2 - 1; voice < NUM_VOICES; --voice){
        phases[voice * 2 + 1] = phases[voice];
    }
}

void Synth::process(float* output, const uint32_t frames)
{
    double hz;

    //Render.
    for(uint32_t i = 0; i < frames; i++){

        //Set pitch.
        hz = hzFund = spectrumPhase * fMin + fMin;

        bool isOddHarmonic = true;
        
        for(uint voice = 0; voice < NUM_VOICES; ++voice){
            //Get new phase.
            double phase = phases[voice];
            phase += hz * sampleInterval;
            phase = frac(phase);

            //Anti-aliasing: don't bother rendering anything over the Nyquist rate.
            if( hz > hzNyq ) break;

            output[i] += (!isOddHarmonic + isOddHarmonic * spectrumPhase * spectrumPhase)   //Fade in odd harmonics.
            * getAmplitude(hz)                                              //Frequency response.
            * sin(2.0 * M_PI * phase);                                      //Additives.

            //Remember phase, move to higher overtone.
            phases[voice] = phase;
            hz *= (voice + 2.0) / (voice + 1.0);
            isOddHarmonic = !isOddHarmonic;
        }

        output[i] *= volume / static_cast<double>(NUM_VOICES);

        //Wrapping.
        spectrumPhase += spectrumPhaseVelocity * sampleInterval;
        if( spectrumPhase > 1.0) shiftDown();
        if( spectrumPhase < 0.0)   shiftUp();
    }

}

void Synth::setSampleRate(double oldRate, double newRate){
    sampleInterval = 1.0 / newRate;
    hzNyq = newRate / 2;
}

//Takes value from 0 to 1 representing frequency shift factor.
//0 : lower one octave per second
//1 : raise one octave per second
void Synth::setPhaseVelocity(double in){
    spectrumPhaseVelocity = 2.0 * (in - 0.5);
}

void Synth::setTimbre(double in){
    
}

void Synth::setVolume(double in){
    volume = in;
}